# CheapBleed release 6 by David Bowland
# ./addons/eventscripts/cheapbleed/cheapbleed.py

# >>> To configure this addon please see cheapbleed.cfg <<<

"""
Makes players lose health and emit blood as well as a pain sound every few seconds after being hurt. Requires only ES 2.0+
"""


import es
import gamethread
import os.path
import random

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'CheapBleed'
info.version  = '6'
info.url      = 'http://addons.eventscripts.com/addons/view/cheapbleed'
info.basename = 'cheapbleed'
info.author   = 'SuperDave'


dict_events  = {'player_bleed':{'userid':'short', 'dmg_health':'short', 'attacker':'short'}}
dict_options = {'cheapbleed_damage_initial':[30, 'Percent of damage taken a player will lose on the first bleed'], 'cheapbleed_damage_attenuation':[45, 'Percent of damage taken by previous bleed that will be lost on the next bleed'], 'cheapbleed_delay_initial':[3, 'Number of seconds between taking damage and the first bleed'], 'cheapbleed_delay_attenuation':[150, 'Percent of previous bleed delay that will be the delay for the next bleed'], 'cheapbleed_sound_percent':[10, 'Percent of the map around a player that can hear when that player bleeds--set to 0 for no sound']}


bleed_sounds = []
players      = {}


class BleedManager:
   userid     = 0
   bleeds     = {}
   last_sound = ''

   def __init__(self, userid):
      self.userid = userid
      self.bleeds = {}

   def newBleed(self, amount, attacker):
      """Creates a new bleed delay for the player"""
      bleed_index = (sorted(self.bleeds)[len(self.bleeds) - 1] + 1) if self.bleeds else 0
      self.bleeds[bleed_index] = attacker

      delay = float(es.ServerVar('cheapbleed_delay_initial'))
      gamethread.delayedname(delay, 'cheapbleed_%s_%s' % (self.userid, bleed_index), self._executeBleed, (bleed_index, amount, delay, attacker))

   def removeDelays(self, attacker=0):
      """Removes bleeds by attacker or in entirety"""
      for bleed_index in filter(lambda x: self.bleeds[x] == attacker, self.bleeds) if attacker else self.bleeds.keys():
         self._removeBleed(bleed_index)

   def _executeBleed(self, bleed_index, amount, delay, attacker):
      """
      Fires player_bleed
      Emits blood and sound
      Calculates the next bleed damage and delay and starts a new bleed delay if necessary
      """
      if not es.exists('userid', self.userid) or not es.exists('userid', attacker):
         self._removeBleed(bleed_index, False)
         return

      es.event('initialize', 'player_bleed')
      es.event('setint', 'player_bleed', 'userid', self.userid)
      es.event('setint', 'player_bleed', 'dmg_health', int(amount))
      es.event('setint', 'player_bleed', 'attacker', attacker)
      es.event('fire', 'player_bleed')

      self._emitBlood(amount, attacker)
      self._emitSound()

      amount = (amount * es.ServerVar('cheapbleed_damage_attenuation')) / 100
      delay  = (delay  * es.ServerVar('cheapbleed_delay_attenuation'))  / 100
      if int(amount) and int(delay):
         gamethread.delayedname(delay, 'cheapbleed_%s_%s' % (self.userid, bleed_index), self._executeBleed, (bleed_index, amount, delay, attacker))

      else:
         self._removeBleed(bleed_index, False)

   def _emitBlood(self, amount, attacker):
      """
      Gives the player the targetname "cheapbleeder"
      Emits blood from the player
      Hurts the player for the required damage
      Gives the player the targetname "none"
      """
      es.server.queuecmd('es_xfire %s !self addoutput \"targetname cheapbleeder\"' % self.userid)

      es.server.queuecmd('es_xfire %s env_blood kill' % self.userid)
      es.server.queuecmd('es_xentcreate %s env_blood' % self.userid)
      es.server.queuecmd('es_xfire %s env_blood addoutput \"parentname cheapbleeder\"' % self.userid)
      es.server.queuecmd('es_xfire %s env_blood addoutput \"amount %s\"' % (self.userid, int(amount) * 10))
      es.server.queuecmd('es_xfire %s env_blood addoutput \"spawnflags 13\"' % self.userid)
      es.server.queuecmd('es_xfire %s env_blood emitblood' % self.userid)

      es.server.queuecmd('es_xfire %s point_hurt kill' % self.userid)
      es.server.queuecmd('es_xentcreate %s point_hurt' % self.userid)
      es.server.queuecmd('es_xfire %s point_hurt addoutput \"damagetarget cheapbleeder\"' % self.userid)
      es.server.queuecmd('es_xfire %s point_hurt addoutput \"damage %s\"' % (self.userid, int(amount)))
      es.server.queuecmd('es_xfire %s point_hurt addoutput \"damagetype 32\"' % self.userid)
      es.server.queuecmd('es_xfire %s point_hurt hurt' % attacker)

      es.server.queuecmd('es_xfire %s !self addoutput \"targetname none\"' % self.userid)

   def _emitSound(self):
      """Emits a bleed sound from the player if necessary"""
      attenuation = float(es.ServerVar('cheapbleed_sound_percent'))
      if attenuation and bleed_sounds:
         possible_sounds = filter(lambda x: x <> self.last_sound, bleed_sounds)
         if not possible_sounds:
            possible_sounds = bleed_sounds[:]

         self.last_sound = random.choice(possible_sounds)
         es.emitsound('player', self.userid, self.last_sound, 1.0, (100 - attenuation) / 100.)

   def _removeBleed(self, bleed_index, cancel_delayed=True):
      """Removes bleed by index"""
      if cancel_delayed:
         gamethread.cancelDelayed('cheapbleed_%s_%s' % (self.userid, bleed_index))

      if self.bleeds.has_key(bleed_index):
         del self.bleeds[bleed_index]


def getPlayer(userid):
   """Returns the BleedManager instance for the userid"""
   global players

   userid = int(userid)
   if not players.has_key(userid):
      players[userid] = BleedManager(userid)

   return players[userid]


def removePlayer(userid, attacker=False):
   """
   Removes BleedManager instance by userid
   Removes all bleed delays involving userid if necessary
   """
   global players

   userid = int(userid)
   if players.has_key(userid):
      players[userid].removeDelays()
      del players[userid]

      if attacker:
         for loop_userid in players:
            players[loop_userid].removeDelays(userid)


def clearDelays():
   """Clears outstanding bleed delays"""
   for userid in players:
      players[userid].removeDelays()

   players.clear()


def load():
   """Checks to make sure critical server variables are created by the config"""

   for str_option in dict_options:
      es.ServerVar(str_option, dict_options[str_option][0], dict_options[str_option][1])
   if os.path.isfile(es.getAddonPath('cheapbleed') + '/cheapbleed.cfg'):
      es.server.cmd('es_xmexec ../addons/eventscripts/cheapbleed/cheapbleed.cfg')
   else:
      es.dbgmsg(0, 'CheapBleed: Unable to load cheapbleed.cfg! Please ensure it is in the ./cheapbleed/ directory.')

   es_map_start({})


def es_map_start(event_var):
   """
   Calls clearDelays to clear delays and prevent bleeding from past maps
   Marks sounds for download
   Calls load_events
   """
   clearDelays()

   path = str(es.ServerVar('eventscripts_gamedir')) + '/sound/'
   for sound in filter(lambda x: os.path.isfile(path + x), bleed_sounds):
      es.stringtable('downloadables', 'sound/' + sound)

   load_events()


def round_start(event_var):
   """Calls clearDelays to clear delays and prevent bleeding from past rounds"""
   clearDelays()


def player_spawn(event_var):
   """Removes bleed delays where the spawning player is the victim"""
   removePlayer(event_var['userid'])


def player_hurt(event_var):
   """Calculates the bleed damage from the hit and starts a bleed delay if that damage is over zero"""
   if event_var['weapon'] <> 'point_hurt':
      damage = (float(event_var['dmg_health']) * es.ServerVar('cheapbleed_damage_initial')) / 100
      if int(damage):
         userid   = int(event_var['userid'])
         attacker = int(event_var['attacker'])

         getPlayer(userid).newBleed(damage, attacker if attacker else userid)


def player_death(event_var):
   """Removes bleed delays where the dead player is the victim"""
   removePlayer(event_var['userid'])


def player_disconnect(event_var):
   """Removes bleed delays where the disconnecting player is the victim or attacker"""
   removePlayer(event_var['userid'], True)


def unload():
   """Calls clearDelays to clear delays"""
   clearDelays()

   for str_option in dict_options:
      es.ServerVar(str_option).set(0)


def addsound_cmd():
   """
   cheapbleed_addsound <"sound relative to ./sound/">
   Adds bleed sound to database
   """
   if es.getargc() == 2:
      bleed_sounds.append(es.getargv(1))

   else:
      es.dbgmsg(0, 'Syntax: cheapbleed_addsound <\"sound relative to ./sound/\">')
if not es.exists('command', 'cheapbleed_addsound'):  
   es.regcmd('cheapbleed_addsound', 'cheapbleed/addsound_cmd', 'cheapbleed_addsound <\"sound relative to ./sound/\">\nAdds a bleed sound')


def removesound_cmd():
   """
   cheapbleed_removesound <"sound relative to ./sound/">
   Removes bleed sound from database
   """
   if es.getargc() == 2:
      sound = es.getargv(1)
      if sound in bleed_sounds: 
         bleed_sounds.remove(sound)

   else:
      es.dbgmsg(0, 'Syntax: cheapbleed_removesound <\"sound relative to ./sound/\">')
if not es.exists('command', 'cheapbleed_removesound'):  
   es.regcmd('cheapbleed_removesound', 'cheapbleed/removesound_cmd', 'cheapbleed_removesound <\"sound relative to ./sound/\">\Removes a bleed sound')


def load_events():
   """Loads the .res file for the events"""
   if os.path.isfile(es.getAddonPath('cheapbleed') + '/cheapbleed.res'):
      es.loadevents('declare', 'addons/eventscripts/cheapbleed/cheapbleed.res')

   else:
      if not os.path.isfile(es.getAddonPath('cheapbleed') + 'es_cheapbleed_events_db.txt'):
         es.keygroupcreate('cheapbleed_events')
         for str_event in dict_events:
            es.keycreate('cheapbleed_events', str_event)
            for str_event_var in dict_events[str_event]:
               es.keysetvalue('cheapbleed_events', str_event, str_event_var, dict_events[str_event][str_event_var])
         es.keygroupsave('cheapbleed_events', '|cheapbleed')
         es.keygroupdelete('cheapbleed_events')
      es.loadevents('declare', 'addons/eventscripts/cheapbleed/es_cheapbleed_events_db.txt')
      es.dbgmsg(0, 'CheapBleed: Unable to load cheapbleed.res! Please ensure it is in the ./cheapbleed/ directory.')